package be.iminds.ilabt.jfed.highlevel.jobs.states;

import be.iminds.ilabt.jfed.espec.ESpecLogic;
import be.iminds.ilabt.jfed.espec.model.*;
import be.iminds.ilabt.jfed.espec.parser.ESpecConstants;
import be.iminds.ilabt.jfed.espec.parser.ESpecDefaults;
import be.iminds.ilabt.jfed.experiment.Experiment;
import be.iminds.ilabt.jfed.highlevel.jobs.SetupSoftwareExperimentJob;
import be.iminds.ilabt.jfed.highlevel.util.ExternalFileUtil;
import be.iminds.ilabt.jfed.lowlevel.connection.JFedException;
import be.iminds.ilabt.jfed.rspec.model.ModelRspec;
import be.iminds.ilabt.jfed.rspec.model.RspecNode;
import javafx.util.Pair;
import net.schmizz.sshj.SSHClient;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

public class RemotePathResolver {
    private static final Logger LOG = LoggerFactory.getLogger(RemotePathResolver.class);

    @Nonnull
    private final SetupSoftwareExperimentJob setupSoftwareExperimentJob;
    @Nonnull
    private final Experiment experiment;
    @Nonnull
    private final ExperimentSpecification experimentSpecification;

    @Nonnull
    private final Map<String, ExternalFileUtil> externalFileUtilByClientId;

    @Nullable
    private boolean resolvedPaths;

    public RemotePathResolver(@Nonnull SetupSoftwareExperimentJob job) {
        this(job, job.getExperiment(), job.getExperiment().getExperimentSpecification());
    }
    public RemotePathResolver(@Nonnull SetupSoftwareExperimentJob setupSoftwareExperimentJob,
                              @Nonnull Experiment experiment,
                              @Nonnull ExperimentSpecification experimentSpecification) {
        this.externalFileUtilByClientId = new HashMap<>();
        this.resolvedPaths = false;

        this.setupSoftwareExperimentJob = setupSoftwareExperimentJob;
        this.experiment = experiment;
        this.experimentSpecification = experimentSpecification;
    }

    private ExternalFileUtil getExternalFileUtil(@Nonnull final RspecNode node) throws JFedException, InterruptedException {
        String clientId = node.getClientId();
        ExternalFileUtil res = externalFileUtilByClientId.get(clientId);
        if (res != null)
            return res;

        SSHClient ssh;
        try {
            ssh = setupSoftwareExperimentJob.getSSHClient(node, null);
        } catch (SetupSoftwareExperimentJob.SshException e) {
            LOG.warn("Aborting in RemotePathResolver due to SSH connection problem", e);
            throw new RuntimeException("Aborting in RemotePathResolver due to SSH connection problem");
        }
        assert ssh != null;
        res = new ExternalFileUtil(ssh);
        return res;
    }

    private final ConcurrentHashMap<Pair<String, ESpecConstants.DirContent>, String> remotePathByDirClientIdAndDirContent = new ConcurrentHashMap<>();
    private final ConcurrentHashMap<Pair<String, UploadLikeSpec>, String> remotePathByClientIdAndUploadLikeSpec = new ConcurrentHashMap<>();
    private final ConcurrentHashMap<Pair<String, UploadLikeWithLogSpec>, String> remoteLogPathByClientIdAndUploadLikeWithLogSpec = new ConcurrentHashMap<>();

    @Nullable
    public String getRemotePath(@Nonnull ESpecConstants.DirContent dirContent, @Nonnull String clientId) throws JFedException, InterruptedException {
        resolvePathsIfNeeded(clientId);
        return remotePathByDirClientIdAndDirContent.get(new Pair<>(clientId, dirContent));
    }
    @Nullable
    public String getRemotePath(@Nonnull DirSpec dirSpec, @Nonnull String clientId) throws JFedException, InterruptedException {
        if (dirSpec.getDirContent() == null)
            return null;
        return this.getRemotePath(dirSpec.getDirContent(), clientId);
    }
    @Nullable
    public String getRemotePath(@Nonnull UploadLikeSpec uploadLikeSpec, @Nonnull String clientId) throws JFedException, InterruptedException {
        resolvePathsIfNeeded(clientId);
        return remotePathByClientIdAndUploadLikeSpec.get(new Pair<>(clientId, uploadLikeSpec));
    }
    @Nullable
    public String getRemoteLogPath(@Nonnull UploadLikeWithLogSpec spec, @Nonnull String clientId) throws JFedException, InterruptedException {
        resolvePathsIfNeeded(clientId);
        return remoteLogPathByClientIdAndUploadLikeWithLogSpec.get(new Pair<>(clientId, spec));
    }



    private String dirPathToAbsolutePath(ExternalFileUtil externalFileUtil, String dirPath) {
        if (externalFileUtil == null)
            throw new IllegalStateException("externalFileUtil == null");

        //3 cases for path:
        //  - absolute path (starts with /)   example: /work/data/ / /work /work/test/data
        //  - home path (starts with ~)       exammple: ~/data/  ~/   ~   ~/test/data
        //  - null -> home path               -> this is handled by the parser, so no need to check this here
        //TODO    empty path string or path not starting with / or ~  ->  error during parsing

        boolean absolutePath = dirPath.startsWith("/");
        assert absolutePath || dirPath.startsWith("~");

        if (absolutePath)
            return assureEndWithSlash(dirPath);

        String homeDir = assureEndWithSlash(externalFileUtil.getHomeDir());

        return assureEndWithSlash(homeDir + (dirPath.length() > 1 ? dirPath.substring(1) : ""));
    }


    @Nonnull
    private String makeRelativePathAbsolute(@Nonnull UploadLikeSpec uploadSpec,
                                            @Nonnull ExternalFileUtil externalFileUtil,
                                            @Nullable String uploadDirSpecPath,
                                            @Nullable String scriptDirSpecPath,
                                            @Nullable String ansibleDirSpecPath,
                                            @Nonnull String path) {
        if (ESpecLogic.isAnsibleSubPart(uploadSpec, experimentSpecification)) {
            if (ansibleDirSpecPath != null) {
                return assureEndWithSlash(convertHomeDir(ansibleDirSpecPath, externalFileUtil.getHomeDir())) +
                        path;
            } else {
                return assureEndWithSlash(externalFileUtil.getHomeDir()) + "ansible/"+ path;
            }
        } else if (uploadSpec instanceof UploadSpec) {
            if (uploadDirSpecPath != null) {
                return assureEndWithSlash(convertHomeDir(uploadDirSpecPath, externalFileUtil.getHomeDir())) +
                        path;
            } else {
                return assureEndWithSlash(externalFileUtil.getHomeDir()) + path;
            }
        } else if (uploadSpec instanceof ExecuteSpec) {
            if (scriptDirSpecPath != null) {
                return assureEndWithSlash(convertHomeDir(scriptDirSpecPath, externalFileUtil.getHomeDir())) +
                        path;
            } else {
                return assureEndWithSlash(externalFileUtil.getHomeDir()) + path;
            }
        } else {
            throw new RuntimeException("BUG: uploadSpec of unknown class " + uploadSpec.getClass().getName());
        }
    }
    @Nonnull
    private String makeAnyPathAbsolute(@Nonnull UploadLikeSpec uploadSpec,
                                       @Nonnull ExternalFileUtil externalFileUtil,
                                       @Nullable String uploadDirSpecPath,
                                       @Nullable String scriptDirSpecPath,
                                       @Nullable String ansibleDirSpecPath,
                                       @Nonnull String path) {
        //3 cases for uploadSpec.getSource().getValue():
        //                                             (the file extensions in the examples below are not required at all)
        //  - absolute path (starts with /)   example: /work/data/filename.dat /testfile.txt
        //  - home path (starts with ~)       example: ~/data/filename.dat  ~/filename.txt
        //  - relative path                   example: filename.dat  test/filename.txt
        //       empty or null uploadSpec.getSource().getValue() -> error during parsing

        boolean absolutePath = path.startsWith("/");
        boolean homePath = path.startsWith("~");
        boolean relativePath = !(absolutePath || homePath);

        if (absolutePath) {
            return path;
        }
        if (homePath) {
            return convertHomeDir(path, externalFileUtil.getHomeDir());
        }

        assert relativePath;

        return makeRelativePathAbsolute(uploadSpec,
                externalFileUtil,
                uploadDirSpecPath,
                scriptDirSpecPath,
                ansibleDirSpecPath,
                path);
    }

    @Nonnull
    private String getRemotePath(@Nonnull ExternalFileUtil externalFileUtil,
                                 @Nullable String uploadDirSpecPath,
                                 @Nullable String scriptDirSpecPath,
                                 @Nullable String ansibleDirSpecPath,
                                 @Nonnull UploadLikeSpec uploadSpec) {
        if (uploadSpec.getPath() == null) {
            if (uploadSpec.getSource() == null)
                throw new IllegalStateException("source and path should never both be null");

            String sourceBasename = removeParentDirs(uploadSpec.getSource().getBasename());

            return makeRelativePathAbsolute(uploadSpec,
                    externalFileUtil,
                    uploadDirSpecPath,
                    scriptDirSpecPath,
                    ansibleDirSpecPath,
                    sourceBasename);
        } else {
            return makeAnyPathAbsolute(uploadSpec,
                    externalFileUtil,
                    uploadDirSpecPath,
                    scriptDirSpecPath,
                    ansibleDirSpecPath,
                    uploadSpec.getPath());
        }
    }

    public void resolvePathsIfNeeded(@Nonnull final String clientId) throws JFedException, InterruptedException {
        assert experiment.getSliceOrNull() != null;
        assert experiment.getSlice().getManifestRspec() != null;
        ModelRspec manifestModel =
                experiment.getSlice().getManifestRspec().getModelRspec();
        assert manifestModel != null;
        RspecNode node = manifestModel.getNodeByClientId(clientId);
        resolvePathsIfNeeded(node);
    }
    public void resolvePathsIfNeeded(@Nonnull final RspecNode node) throws JFedException, InterruptedException {
        if (resolvedPaths)
            return;

        assert experiment.getExperimentSpecificationFileManager() != null;

        String clientId = node.getClientId();
        assert clientId != null;

        ExternalFileUtil externalFileUtil = getExternalFileUtil(node);
        assert externalFileUtil != null;

        List<DirSpec> dirs = new ArrayList<>(experimentSpecification.getDirs());

        for (DirSpec dirSpec : dirs) {
            if (ESpecLogic.mustRunOn(clientId, node.getAnsibleGroups(), experimentSpecification, dirSpec)) {
                String remotePath = dirPathToAbsolutePath(externalFileUtil, dirSpec.getPath());
                if (dirSpec.getDirContent() != null) {
                    remotePathByDirClientIdAndDirContent.put(new Pair<>(clientId, dirSpec.getDirContent()), remotePath);
                }
            }
        }

        List<UploadLikeSpec> allUploads = new ArrayList<>(experimentSpecification.getUploads());
        allUploads.addAll(experimentSpecification.getExecutes());
        List<UploadLikeSpec> ansibleUploads = new ArrayList<>();
        if (experimentSpecification.getAnsible() != null) {
            ansibleUploads.addAll(experimentSpecification.getAnsible().getAnsibleHostSpec().getUploadLists());
            ansibleUploads.addAll(experimentSpecification.getAnsible().getAnsibleHostSpec().getExecuteLists());
            ansibleUploads.addAll(experimentSpecification.getAnsible().getAnsibleGalaxySpecs());
            ansibleUploads.addAll(experimentSpecification.getAnsible().getAnsiblePlaybookSpecs());
            for (AnsiblePlaybookSpec ansiblePlaybookSpec : experimentSpecification.getAnsible().getAnsiblePlaybookSpecs()) {
                if (ansiblePlaybookSpec.hasExtraVarsJsonFileSpec() &&
                        ansiblePlaybookSpec.getExtraVarsJsonFileSpec() != null) {
                    ansibleUploads.add(ansiblePlaybookSpec.getExtraVarsJsonFileSpec());
                }
            }
            allUploads.addAll(ansibleUploads);
        }

        String uploadDirSpecPath = null;
        String scriptDirSpecPath = null;
        String ansibleDirSpecPath = null;
        for (DirSpec dirSpec : experimentSpecification.getDirs()) {
            if (dirSpec.getNodes() == null || dirSpec.getNodes().contains(clientId)) {
                if (dirSpec.isScriptContent()) {
                    scriptDirSpecPath = dirPathToAbsolutePath(externalFileUtil, dirSpec.getPath());
                }
                if (dirSpec.isUploadContent()) {
                    uploadDirSpecPath = dirPathToAbsolutePath(externalFileUtil, dirSpec.getPath());
                }
                if (dirSpec.isAnsibleContent()) {
                    ansibleDirSpecPath = dirPathToAbsolutePath(externalFileUtil, dirSpec.getPath());
                }
            }
        }
        assert scriptDirSpecPath != null;
        assert uploadDirSpecPath != null;
//      assert ansibleDirSpecPath != null; //can be null if no ansible


        for (UploadLikeSpec uploadSpec : allUploads) {
            boolean remoteOnlyFile = uploadSpec.getSource() == null;

            if (!ESpecLogic.mustRunOn(clientId, node.getAnsibleGroups(), experimentSpecification, uploadSpec)) {
                //not on this node
                continue;
            }

            //use dir with "scripts" or "uploads" type as appropriate, fall back to homedir
            String remotePath = getRemotePath(externalFileUtil,
                    uploadDirSpecPath, scriptDirSpecPath, ansibleDirSpecPath,
                    uploadSpec);

            String name = ExternalFileUtil.getBasename(remotePath);

            boolean isDir = uploadSpec.getSource() != null && experiment.getExperimentSpecificationFileManager().isDir(uploadSpec.getSource());
            if (isDir) {
                if (!remotePath.endsWith("/"))
                    remotePath += "/";
            }

            remotePathByClientIdAndUploadLikeSpec.put(new Pair<>(clientId, uploadSpec), remotePath);

            if (uploadSpec instanceof UploadLikeWithLogSpec) {
                UploadLikeWithLogSpec es = (UploadLikeWithLogSpec) uploadSpec;

                final String baseLogPath;
                if (es.getLog() != null) {
                    baseLogPath = es.getLog();
                } else {
                    if (experimentSpecification.getConfig().isStoreRemoteLogsByDefault()) {
                        baseLogPath = ESpecDefaults.deriveAutomaticLogFile(remotePath);
                    } else {
                        baseLogPath = null;
                    }
                }

                if (baseLogPath != null) {
                    String remoteLogPath;
                    if (!baseLogPath.contains("/")) {
                        //file is considered to be in the same dir as the executed file
                        String dir = ExternalFileUtil.getPathOnly(remotePath);
                        remoteLogPath = assureEndWithSlash(dir)+baseLogPath;
                    } else {
                        remoteLogPath = convertHomeDir(baseLogPath, externalFileUtil.getHomeDir());
                    }
                    assert remoteLogPath != null;
                    remoteLogPathByClientIdAndUploadLikeWithLogSpec.put(new Pair<>(clientId, es), remoteLogPath);
                }
            }
        }
    }




    @Nonnull
    private static String convertHomeDir(@Nonnull String path, @Nonnull String homeDirPath) {
        if (path.startsWith("~/") && path.length() > 2) return assureEndWithSlash(homeDirPath)+path.substring(2);
        if (path.startsWith("~") && path.length() > 1) return assureEndWithSlash(homeDirPath)+path.substring(1);
        if (path.equals("~")) return assureEndWithSlash(homeDirPath);
        return path;
    }

    @Nonnull
    private static String assureEndWithSlash(@Nonnull String p) {
        if (p.endsWith("/")) return p;
        return p + "/";
    }

    @Nonnull
    private static String assureStartWithSlash(@Nonnull String p) {
        if (p.startsWith("/")) return p;
        return "/" + p;
    }

    @Nonnull
    private static String removeLastChildDir(@Nonnull String childDir, @Nonnull String path) {
        String childDirS = assureEndWithSlash(childDir);
        if (childDir.equals("/")) return path;
        assert path.endsWith(childDirS) : "\""+path+"\" does not end with \""+childDirS+"\"";
        return path.substring(0, path.length()-childDirS.length());
    }
    @Nonnull
    private static String removeParentDir(@Nonnull String parentPath, @Nonnull String path) {
        if (parentPath.isEmpty()) return path;
        assert path.startsWith(parentPath) : "\""+path+"\" does not start with \""+parentPath+"\"";
        return path.substring(parentPath.length());
    }
    @Nonnull
    private static String removeParentDirs(@Nonnull String path) {
        if (path.isEmpty() || path.indexOf('/') == -1 || path.length() <= 2) return path;

        int slash = path.lastIndexOf('/', path.length()-2);
        if (slash == -1)
            return path;
        return path.substring(slash+1, path.length());
    }
}
